@@ -1,10 +1,11 @@
-#include <imgui.h>
 #include <ios>
-#include <misc/cpp/imgui_stdlib.h>
-#include <pybind11/cast.h>
-#include <pybind11/detail/common.h>
 #include <regex>
+#include <sstream>
+#include <iomanip>
+#include <stdexcept>
+#include <filesystem>
 #include <functional>
+#include <unordered_map>
 
 #include <GL/glew.h>
 #include <GLFW/glfw3.h>
@@ -14,19 +15,25 @@
 #include <pybind11/pytypes.h>
 #include <pybind11/pybind11.h>
 
-#include <experimental/filesystem>
-#include <sstream>
-#include <iomanip>
-#include <stdexcept>
-#include <unordered_map>
-
 #include "backends/imgui_impl_glfw.h"
 #include "backends/imgui_impl_opengl3.h"
-#include "implot.h"
+
+#include "imgui.h"
+
+#ifndef IMGUI_DEFINE_MATH_OPERATORS
+#define IMGUI_DEFINE_MATH_OPERATORS
+#endif
+
 #include "imgui_internal.h"
+#include "misc/cpp/imgui_stdlib.h"
+
+#include "implot.h"
 
 #include "SourceSansPro.h"
+#include "input.h"
 
+
+namespace fs = std::filesystem;
 namespace py = pybind11;
 
 struct ImViz {
@@ -38,9 +45,9 @@ struct ImViz {
     bool mod = false;
     bool mod_any = false;
 
-    std::regex re{"(-)?(o|s|d|\\*|\\+)?(r|g|b|y|m|w)?"};
+    std::regex re{"(-)?(o|s|d|\\*|\\+)?"};
 
-    std::unordered_map<std::string, GLuint> textureCache;
+    std::unordered_map<ImGuiID, GLuint> textureCache;
 
     ImViz () {
 
@@ -70,20 +77,19 @@ struct ImViz {
             std::exit(-1);
         }
 
+        // input setup
+        
+        input::registerCallbacks(window);
+
         // basic imgui setup
 
         IMGUI_CHECKVERSION();
         ImGui::CreateContext();
         ImPlot::CreateContext();
 
-        ImGui_ImplGlfw_InitForOpenGL(window, false);
+        ImGui_ImplGlfw_InitForOpenGL(window, true);
         ImGui_ImplOpenGL3_Init("#version 330");
 
-        glfwSetKeyCallback(window, ImGui_ImplGlfw_KeyCallback);
-        glfwSetCharCallback(window, ImGui_ImplGlfw_CharCallback);
-        glfwSetScrollCallback(window, ImGui_ImplGlfw_ScrollCallback);
-        glfwSetMouseButtonCallback(window, ImGui_ImplGlfw_MouseButtonCallback);
-
         ImGuiIO& io = ImGui::GetIO();
         io.ConfigFlags |= ImGuiConfigFlags_DockingEnable;
         io.IniFilename = "./imviz.ini";
@@ -102,44 +108,22 @@ struct ImViz {
 
     void prepareUpdate() {
 
+        input::update();
+
         ImGuiIO& io = ImGui::GetIO();
 
         if (io.WantCaptureMouse) {
-            // do something ... 
+            input::clearMouseInput();
         }
+
         if (io.WantCaptureKeyboard) {
-            // do something ... 
+            input::clearKeyboardInput();
         }
 
         ImGui_ImplOpenGL3_NewFrame();
         ImGui_ImplGlfw_NewFrame();
         ImGui::NewFrame();
 
-        static bool showLightTheme = false;
-        static bool showImGuiDemo = false;
-        static bool showImPlotDemo = false;
-
-        if (ImGui::BeginMainMenuBar()) {
-
-            if (ImGui::BeginMenu("Show")) {
-
-                ImGui::MenuItem("Light Theme", NULL, &showLightTheme);
-                ImGui::MenuItem("ImGui Demo", NULL, &showImGuiDemo);
-                ImGui::MenuItem("ImPlot Demo", NULL, &showImPlotDemo);
-                ImGui::EndMenu();
-            }
-
-            ImGui::EndMainMenuBar();
-        }
-
-        if (showLightTheme) {
-            ImGui::StyleColorsLight();
-            ImPlot::StyleColorsLight();
-        } else {
-            ImGui::StyleColorsDark();
-            ImPlot::StyleColorsDark();
-        }
-
         // dock space
 
         ImGuiWindowFlags window_flags = ImGuiWindowFlags_NoDocking
@@ -172,14 +156,6 @@ struct ImViz {
         ImGui::DockSpace(1, ImVec2(0.0f, 0.0f), dockspace_flags);
 
         ImGui::End();
-
-        if (showImGuiDemo) {
-            ImGui::ShowDemoWindow(&showImGuiDemo);
-        }
-
-        if (showImPlotDemo) {
-            ImPlot::ShowDemoWindow(&showImPlotDemo);
-        }
     }
 
     void doUpdate (bool useVsync) {
@@ -223,9 +199,149 @@ struct ImViz {
         mod = m;
         mod_any |= m;
     }
-
 };
 
+/**
+ * Custom ImGui Extension for handling path selection.
+ */
+namespace ImGui {
+
+    void PathSelector (std::string& selectedPath) { 
+
+        ImGui::BeginChild("Dir Listing",
+                ImVec2(500, 400),
+                false,
+                ImGuiWindowFlags_HorizontalScrollbar);
+
+        if (ImGui::Selectable("./..", false)) {
+            if (!fs::is_directory(selectedPath) 
+                    || fs::path(selectedPath).filename().empty()) {
+                selectedPath = fs::path(selectedPath)
+                    .parent_path()
+                    .parent_path();
+            } else {
+                selectedPath = fs::path(selectedPath)
+                    .parent_path();
+            }
+        }
+
+        // Obtain a list of all entries in the current directory
+
+        fs::path listPath = selectedPath;
+        if (!fs::is_directory(listPath)) {
+            listPath = listPath.parent_path();
+        }
+
+        std::vector<fs::directory_entry> entries;
+
+        if (fs::exists(listPath)) {
+            for (fs::directory_entry e : fs::directory_iterator(listPath)) {
+                if (std::string(e.path().filename()).at(0) != '.') {
+                    entries.push_back(e);
+                }
+            }
+        }
+
+        // Because the entries may be sorted arbitrarily
+        // we sort them alphabetically.
+        // Also directories are sorted in before files.
+
+        std::sort(
+                entries.begin(),
+                entries.end(),
+                [](auto& a, auto& b) -> bool {
+                    if (fs::is_directory(a) && !fs::is_directory(b)) {
+                        return true;
+                    } else if (!fs::is_directory(a) && fs::is_directory(b)) {
+                        return false;
+                    } else {
+                        return a.path().filename() < b.path().filename();
+                    }
+                }
+            );
+
+        for (fs::directory_entry e : entries) {
+
+            std::string displayName = e.path().stem();
+            displayName += e.path().extension();
+
+            if (fs::is_directory(e)) {
+                displayName += "/";
+            }
+
+            if (ImGui::Selectable(
+                        displayName.c_str(),
+                        selectedPath == e.path())) {
+
+                selectedPath = e.path();
+            }
+        }
+
+        ImGui::EndChild();
+
+        char filenameInputBuf[256];
+
+        strncpy(filenameInputBuf, 
+                selectedPath.c_str(), 
+                sizeof(filenameInputBuf) - 1);
+
+        ImGui::PushItemWidth(ImGui::GetContentRegionAvailWidth() * 0.8f);
+        ImGui::InputText("selected",
+                filenameInputBuf, IM_ARRAYSIZE(filenameInputBuf));
+        ImGui::PopItemWidth();
+
+        selectedPath = std::string(filenameInputBuf);
+    }
+
+    bool FileDialogPopup (
+            const char* name,
+            const char* confirmLabel,
+            std::string& selectedPath) {
+
+        bool result = false;
+
+        ImVec2 center = ImGui::GetMainViewport()->GetCenter();
+        ImGui::SetNextWindowPos(center, ImGuiCond_Appearing, ImVec2(0.5f, 0.5f));
+
+        static std::string currentPath = "";
+        static bool fileDialogOpen = false;
+
+        if (ImGui::BeginPopupModal(
+                    name, 
+                    NULL, 
+                    ImGuiWindowFlags_AlwaysAutoResize)) {
+
+            if (!fileDialogOpen) {
+                currentPath = selectedPath;
+            }
+            fileDialogOpen = true;
+
+            ImGui::PathSelector(currentPath);
+
+            if (ImGui::Button(confirmLabel, ImVec2(120, 0))) {
+                ImGui::CloseCurrentPopup();
+                selectedPath = currentPath;
+                result = true;
+            }
+
+            ImGui::SetItemDefaultFocus();
+
+            ImGui::SameLine();
+
+            if (ImGui::Button("Cancel", ImVec2(120, 0))) {
+                ImGui::CloseCurrentPopup();
+                result = false;
+            }
+
+            ImGui::EndPopup();
+        } else {
+            fileDialogOpen = false;
+        }
+
+        return result;
+    }
+}
+
 /**
  * Some helpers to make handling arrays easier.
  */
@@ -233,6 +349,22 @@ struct ImViz {
 template <typename T>
 using array_like = py::array_t<T, py::array::c_style | py::array::forcecast>;
 
+std::string shapeToStr(py::array& array) {
+
+    std::stringstream ss;
+    ss << "(";
+
+    for (int k = 0; k < array.ndim(); ++k) {
+        ss << array.shape()[k];
+        if (k != array.ndim() - 1) {
+            ss <<  ", ";
+        }
+    }
+    ss << ")";
+
+    return ss.str();
+}
+
 void assertArrayShape(std::string name,
                       py::array& array,
                       std::vector<std::vector<int>> shapes) {
@@ -258,7 +390,7 @@ void assertArrayShape(std::string name,
 
     if (!foundShape) {
         std::stringstream ss;
-        ss << "Expected shape of \"" + name + "\" to fit ";
+        ss << "Expected \"" + name + "\" with shape ";
 
         for (size_t i = 0; i < shapes.size(); ++i) { 
             ss << "(";
@@ -275,15 +407,7 @@ void assertArrayShape(std::string name,
             }
         }
 
-        ss << ", but found (";
-
-        for (int k = 0; k < array.ndim(); ++k) { 
-            ss << array.shape()[k]; 
-            if (k != array.ndim() - 1) { 
-                ss <<  ", ";
-            }
-        }
-        ss << ")";
+        ss << ", but found " << shapeToStr(array) << std::endl;
 
         throw std::runtime_error(ss.str());
     }
@@ -323,53 +447,56 @@ struct ImageInfo {
 
     int imageWidth = 0;
     int imageHeight = 0;
-    GLuint textureId = 0;
+    int channels = 0;
+    GLenum format = 0;
+    GLenum datatype = 0;
 };
 
-ImageInfo processImage(std::string& id, py::array& image) {
+ImageInfo interpretImage(py::array& image) {
 
     assert_shape(image, {{-1, -1}, {-1, -1, 1}, {-1, -1, 3}, {-1, -1, 4}});
 
     // determine image parameters
-
-    int imageWidth = 0;
-    int imageHeight = 0;
-    int channels = 0;
-    GLenum format = 0;
-    GLenum datatype = 0;
+    
+    ImageInfo i;
 
     if (image.ndim() == 2) {
-        imageWidth = image.shape(1);
-        imageHeight = image.shape(0);
-        channels = 1;
+        i.imageWidth = image.shape(1);
+        i.imageHeight = image.shape(0);
+        i.channels = 1;
     } else if (image.ndim() == 3) {
-        imageWidth = image.shape(1);
-        imageHeight = image.shape(0);
-        channels = image.shape(2);
+        i.imageWidth = image.shape(1);
+        i.imageHeight = image.shape(0);
+        i.channels = image.shape(2);
     }
 
-    if (channels == 1) {
-        format = GL_RED;
-    } else if (channels == 3) {
-        format = GL_RGB;
-    } else if (channels == 4) {
-        format = GL_RGBA;
+    if (i.channels == 1) {
+        i.format = GL_RED;
+    } else if (i.channels == 3) {
+        i.format = GL_RGB;
+    } else if (i.channels == 4) {
+        i.format = GL_RGBA;
     } 
 
     if (py::str(image.dtype()).equal(py::str("uint8"))) {
-        datatype = GL_UNSIGNED_BYTE;
+        i.datatype = GL_UNSIGNED_BYTE;
     } else if (py::str(image.dtype()).equal(py::str("float32"))) {
-        datatype = GL_FLOAT;
+        i.datatype = GL_FLOAT;
     } else {
-        throw std::runtime_error(
-                "Pixel datatype "
-                + std::string(py::str(image.dtype()))
-                + " not supported. "
-                + "Supported datatypes are uint8 and float32.");
+        i.datatype = GL_FLOAT;
+        image = array_like<float>::ensure(image);
     }
 
+    return i;
+}
+
+GLuint uploadImage(std::string id, ImageInfo& i, py::array& image) {
+
+    ImGuiID uniqueId = ImGui::GetID(id.c_str());
+
     // create a texture if necessary
-    if (viz.textureCache.find(id) == viz.textureCache.end()) {
+    
+    if (viz.textureCache.find(uniqueId) == viz.textureCache.end()) {
 
         GLuint textureId;
 
@@ -384,16 +511,16 @@ ImageInfo processImage(std::string& id, py::array& image) {
 
         glBindTexture(GL_TEXTURE_2D, 0);
 
-        viz.textureCache[id] = textureId;
+        viz.textureCache[uniqueId] = textureId;
     }
 
     // upload texture
 
-    GLuint textureId = viz.textureCache[id];
+    GLuint textureId = viz.textureCache[uniqueId];
 
     glBindTexture(GL_TEXTURE_2D, textureId);
 
-    if (format == GL_RED) {
+    if (i.format == GL_RED) {
         GLint swizzleMask[] = {GL_RED, GL_RED, GL_RED, GL_ONE};
         glTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_SWIZZLE_RGBA, swizzleMask);
     }
@@ -401,28 +528,253 @@ ImageInfo processImage(std::string& id, py::array& image) {
     glTexImage2D(
             GL_TEXTURE_2D,
             0,
-            format,
-            imageWidth,
-            imageHeight,
+            i.format,
+            i.imageWidth,
+            i.imageHeight,
             0,
-            format,
-            datatype,
+            i.format,
+            i.datatype,
             image.data());
 
     glBindTexture(GL_TEXTURE_2D, 0);
 
-    // return info
+    return textureId;
+}
 
-    ImageInfo info;
-    info.imageWidth = imageWidth;
-    info.imageHeight = imageHeight;
-    info.textureId = textureId;
+struct PlotArrayInfo {
+
+    std::vector<double> indices;
+    const double* xDataPtr = nullptr;
+    const double* yDataPtr = nullptr;
+    size_t count = 0;
+};
+
+PlotArrayInfo interpretPlotArrays(
+        array_like<double>& x,
+        array_like<double>& y) {
+
+    PlotArrayInfo info;
+
+    size_t yCount = y.shape()[0];
+
+    if (1 == x.ndim() && 0 == yCount) {
+        // one 1d array given
+        // assume x is [0, 1, 2, ..., N]
+        info.count = x.shape()[0];
+        info.indices.resize(info.count);
+        for (size_t i = 0; i < info.count; ++i) {
+            info.indices[i] = i;
+        }
+        info.xDataPtr = info.indices.data();
+        info.yDataPtr = x.data();
+    } else if (2 == x.ndim() && 0 == yCount) {
+        // one 2d array given
+        size_t len0 = x.shape()[0];
+        size_t len1 = x.shape()[1];
+        if (len0 == 2) {
+            info.xDataPtr = x.data();
+            info.yDataPtr = x.data() + len1;
+            info.count = len1;
+        }
+    } else if (1 == x.ndim() && 1 == y.ndim()) {
+        // two 2d arrays given
+        info.count = std::min(x.shape()[0], y.shape()[0]);
+        info.xDataPtr = x.data();
+        info.yDataPtr = y.data();
+    }
+
+    if (info.count == 0) {
+        throw std::runtime_error(
+                "Plot data with x-shape "
+                + shapeToStr(x)
+                + " and y-shape "
+                + shapeToStr(y)
+                + " cannot be interpreted");
+    }
 
     return info;
 }
 
+/*
+ * Custom type-casters
+ */
+namespace pybind11 {
+    namespace detail {
+
+        /**
+         * ImGui/ImPlot datatypes to numpy-array-like objects
+         */
+
+        template<> struct type_caster<ImVec2> {
+
+        public:
+            PYBIND11_TYPE_CASTER(ImVec2, _("ImVec2"));
+
+            bool load(handle src, bool) {
+
+                auto array = array_like<float>::ensure(src);
+
+                assert_shape(array, {{2,}});
+
+                value.x = array.at(0);
+                value.y = array.at(1);
+
+                return true;
+            }
+
+            static handle cast(
+                    const ImVec2& src,
+                    return_value_policy policy,
+                    handle parent) {
+
+                if (return_value_policy::copy == policy) {
+                    pybind11::array_t<float> array(2, (float*)(&src));
+                    return array.release();
+                } else {
+                    pybind11::array_t<float> array(2, (float*)(&src), parent);
+                    return array.release();
+                }
+            }
+        };
+
+        template<> struct type_caster<ImPlotPoint> {
+
+        public:
+            PYBIND11_TYPE_CASTER(ImPlotPoint, _("ImPlotPoint"));
+
+            bool load(handle src, bool) {
+
+                auto array = array_like<double>::ensure(src);
+
+                assert_shape(array, {{2,}});
+
+                value.x = array.at(0);
+                value.y = array.at(1);
+
+                return true;
+            }
+
+            static handle cast(
+                    const ImPlotPoint& src,
+                    return_value_policy policy,
+                    handle parent) {
+
+                if (return_value_policy::copy == policy) {
+                    pybind11::array_t<double> array(2, (double*)(&src));
+                    return array.release();
+                } else {
+                    pybind11::array_t<double> array(2, (double*)(&src), parent);
+                    return array.release();
+                }
+            }
+        };
+
+        template<> struct type_caster<ImPlotRange> {
+
+        public:
+            PYBIND11_TYPE_CASTER(ImPlotRange, _("ImPlotRange"));
+
+            bool load(handle src, bool) {
+
+                auto array = array_like<double>::ensure(src);
+
+                assert_shape(array, {{2,}});
+
+                value.Min = array.at(0);
+                value.Max = array.at(1);
+
+                return true;
+            }
+
+            static handle cast(
+                    const ImPlotRange& src,
+                    return_value_policy policy,
+                    handle parent) {
+
+                if (return_value_policy::copy == policy) {
+                    pybind11::array_t<double> array(2, (double*)(&src));
+                    return array.release();
+                } else {
+                    pybind11::array_t<double> array(2, (double*)(&src), parent);
+                    return array.release();
+                }
+            }
+        };
+
+        template<> struct type_caster<ImVec4> {
+
+        public:
+            PYBIND11_TYPE_CASTER(ImVec4, _("ImVec4"));
+
+            bool load(handle src, bool) {
+
+                auto array = array_like<float>::ensure(src);
+
+                assert_shape(array, {{4,}});
+
+                value.x = array.at(0);
+                value.y = array.at(1);
+                value.z = array.at(2);
+                value.w = array.at(3);
+
+                return true;
+            }
+
+            static handle cast(
+                    const ImVec4& src,
+                    return_value_policy policy,
+                    handle parent) {
+
+                if (return_value_policy::copy == policy) {
+                    pybind11::array_t<float> array(4, (float*)(&src));
+                    return array.release();
+                } else {
+                    pybind11::array_t<float> array(4, (float*)(&src), parent);
+                    return array.release();
+                }
+            }
+        };
+
+        template<> struct type_caster<ImPlotLimits> {
+
+        public:
+            PYBIND11_TYPE_CASTER(ImPlotLimits, _("ImPlotLimits"));
+
+            bool load(handle src, bool) {
+
+                auto array = array_like<double>::ensure(src);
+
+                assert_shape(array, {{4,}});
+
+                value.X.Min = array.at(0);
+                value.X.Max = array.at(1);
+                value.Y.Min = array.at(2);
+                value.Y.Max = array.at(3);
+
+                return true;
+            }
+
+            static handle cast(
+                    const ImPlotLimits& src,
+                    return_value_policy policy,
+                    handle parent) {
+
+                if (return_value_policy::copy == policy) {
+                    pybind11::array_t<double> array(4, (double*)(&src));
+                    return array.release();
+                } else {
+                    pybind11::array_t<double> array(4, (double*)(&src), parent);
+                    return array.release();
+                }
+            }
+        };
+    }
+}
+
 PYBIND11_MODULE(imviz, m) {
 
+    input::loadPythonBindings(m);
+
     py::enum_<ImGuiCond_>(m, "Cond")
         .value("NONE", ImGuiCond_None)
         .value("ALWAYS", ImGuiCond_Always)
@@ -431,12 +783,68 @@ PYBIND11_MODULE(imviz, m) {
         .value("APPEARING", ImGuiCond_Appearing)
         .export_values();
 
+    m.def("show_imgui_demo", ImGui::ShowDemoWindow);
+    m.def("show_implot_demo", ImPlot::ShowDemoWindow);
+
+    m.def("style_colors_dark", [&](){
+        ImGui::StyleColorsDark();
+        ImPlot::StyleColorsDark();
+    });
+
+    m.def("style_colors_light", [&](){
+        ImGui::StyleColorsLight();
+        ImPlot::StyleColorsLight();
+    });
+
+    m.def("set_main_window_title", [&](std::string title) {
+        glfwSetWindowTitle(viz.window, title.c_str());
+    },
+    py::arg("title"));
+
+    m.def("set_main_window_size", [&](ImVec2 size) {
+        glfwSetWindowSize(viz.window, size.x, size.y);
+    },
+    py::arg("size"));
+
+    m.def("get_main_window_size", [&]() {
+        int w, h;
+        glfwGetWindowSize(viz.window, &w, &h);
+        return ImVec2(w, h);
+    });
+
+    m.def("set_main_window_pos", [&](ImVec2 pos) {
+        glfwSetWindowPos(viz.window, pos.x, pos.y);
+    },
+    py::arg("size"));
+
+    m.def("get_main_window_pos", [&]() {
+        int x, y;
+        glfwGetWindowPos(viz.window, &x, &y);
+        return ImVec2(x, y);
+    });
+
+    m.def("get_main_window_size", [&]() {
+        int w, h;
+        glfwGetWindowSize(viz.window, &w, &h);
+        return ImVec2(w, h);
+    });
+
+    m.def("hide_main_window", [&]() {
+        glfwHideWindow(viz.window);
+    });
+
+    m.def("show_main_window", [&]() {
+        glfwShowWindow(viz.window);
+    });
+
     m.def("mod", [&]() { return viz.mod; });
     m.def("mod_any", [&]() { return viz.mod_any; });
+    m.def("set_mod", [&](bool m) { viz.setMod(m); });
     m.def("clear_mod_any", [&]() { viz.mod_any = false; });
 
     m.def("wait", [&](bool vsync) {
         viz.doUpdate(vsync);
+        input::update();
         glfwPollEvents();
         viz.prepareUpdate();
         return !glfwWindowShouldClose(viz.window);
@@ -447,6 +855,26 @@ PYBIND11_MODULE(imviz, m) {
         viz.trigger();
     });
 
+    m.def("set_next_window_pos",
+            ImGui::SetNextWindowPos,
+    py::arg("position"),
+    py::arg("cond") = ImGuiCond_None,
+    py::arg("pivot") = py::array());
+
+    m.def("set_next_window_size", 
+            ImGui::SetNextWindowSize,
+    py::arg("size"),
+    py::arg("cond") = ImGuiCond_None);
+
+    m.def("set_next_item_width", 
+            ImGui::SetNextItemWidth,
+    py::arg("width"));
+
+    m.def("set_next_item_open", 
+            ImGui::SetNextItemOpen,
+    py::arg("open"),
+    py::arg("cond"));
+ 
     m.def("begin_window", [&](std::string label,
                        bool opened,
                        array_like<float> position, 
@@ -500,17 +928,44 @@ PYBIND11_MODULE(imviz, m) {
 
     m.def("end_window", ImGui::End);
 
-    m.def("window_open", [&]() { return viz.currentWindowOpen; });
+    m.def("get_window_open", [&]() { 
+        return viz.currentWindowOpen;
+    });
+
+    m.def("get_viewport_center", [&]() { 
+        return ImGui::GetMainViewport()->GetCenter();
+    });
+
+    m.def("get_window_pos", ImGui::GetWindowPos);
+    m.def("get_window_size", ImGui::GetWindowSize);
+
+    m.def("get_item_id", ImGui::GetItemID);
+
+    m.def("is_item_focused", ImGui::IsItemFocused);
+    m.def("is_item_active", ImGui::IsItemActive);
+    m.def("is_item_activated", ImGui::IsItemActivated);
+    m.def("is_item_visible", ImGui::IsItemVisible);
+    m.def("is_item_clicked", [&](int mouseButton) {
+        return ImGui::IsItemClicked(mouseButton);
+    },
+    py::arg("mouse_button") = 0);
+
+    m.def("is_item_hovered", [&]() { 
+        return ImGui::IsItemHovered();
+    });
 
     m.def("begin_plot", [&](std::string label,
                             std::string xLabel,
                             std::string yLabel,
+                            array_like<float> size,
                             bool equalAxis,
                             bool autoFitX,
                             bool autoFitY) {
 
         ImPlotFlags flags = 0;
 
+        flags |= ImPlotFlags_Query;
+
         if (equalAxis) {
             flags |= ImPlotFlags_Equal;
         }
@@ -518,6 +973,14 @@ PYBIND11_MODULE(imviz, m) {
         ImPlotAxisFlags xFlags = 0;
         ImPlotAxisFlags yFlags = 0;
 
+        ImVec2 plotSize = ImGui::GetContentRegionAvail();
+
+        if (size.shape()[0] > 0) {
+            assert_shape(size, {{2}});
+            const float* data = size.data();
+            plotSize = ImVec2(data[0], data[1]);
+        } 
+
         if (autoFitX) { 
             xFlags |= ImPlotAxisFlags_AutoFit;
         }
@@ -529,7 +992,7 @@ PYBIND11_MODULE(imviz, m) {
         return ImPlot::BeginPlot(label.c_str(),
                       xLabel.c_str(),
                       yLabel.c_str(),
-                      ImGui::GetContentRegionAvail(),
+                      plotSize,
                       flags,
                       xFlags,
                       yFlags);
@@ -537,17 +1000,24 @@ PYBIND11_MODULE(imviz, m) {
     py::arg("label") = "",
     py::arg("x_label") = "",
     py::arg("y_label") = "",
+    py::arg("size") = py::array_t<float>(),
     py::arg("equal_axis") = false,
     py::arg("auto_fit_x") = false,
     py::arg("auto_fit_y") = false);
 
     m.def("end_plot", &ImPlot::EndPlot);
 
-    m.def("tree_node", [&](std::string label) {
+    m.def("tree_node", [&](std::string label, bool selected) {
 
-        return ImGui::TreeNode(label.c_str());
+        ImGuiTreeNodeFlags flags = ImGuiTreeNodeFlags_None;
+        if (selected) {
+            flags |= ImGuiTreeNodeFlags_Selected;
+        }
+
+        return ImGui::TreeNodeEx(label.c_str(), flags);
     },
-    py::arg("label") = "");
+    py::arg("label") = "",
+    py::arg("selected") = false);
 
     m.def("tree_pop", ImGui::TreePop);
 
@@ -585,6 +1055,10 @@ PYBIND11_MODULE(imviz, m) {
 
         size_t len = items.size();
 
+        if (len == 0) {
+            return selection;
+        }
+
         std::vector<std::string> objStr(len);
         std::vector<const char*> objPtr(len);
 
@@ -606,7 +1080,7 @@ PYBIND11_MODULE(imviz, m) {
         bool mod = ImGui::Combo(label.c_str(), &selectionIndex, objPtr.data(), len);
         viz.setMod(mod);
 
-        return items[selectionIndex];
+        return py::handle(items[selectionIndex]);
     },
     py::arg("label"),
     py::arg("items"),
@@ -678,6 +1152,19 @@ PYBIND11_MODULE(imviz, m) {
     py::arg("min") = 0.0,
     py::arg("max") = 1.0);
 
+    m.def("drag", [&](std::string title, int& value, float speed, int min, int max) {
+        
+        bool mod = ImGui::DragInt(title.c_str(), &value, speed, min, max);
+        viz.setMod(mod);
+
+        return value;
+    }, 
+    py::arg("label"),
+    py::arg("value"),
+    py::arg("speed") = 1.0,
+    py::arg("min") = 0,
+    py::arg("max") = 0);
+
     m.def("drag", [&](std::string title, double& value, float speed, double min, double max) {
         
         bool mod = ImGui::DragScalar(title.c_str(),
@@ -807,9 +1294,11 @@ PYBIND11_MODULE(imviz, m) {
                 std::string id,
                 py::array& image,
                 int displayWidth,
-                int displayHeight) {
+                int displayHeight,
+                array_like<double> tint,
+                array_like<double> borderCol) {
 
-        ImageInfo info = processImage(id, image);
+        ImageInfo info = interpretImage(image);
 
         if (displayWidth < 0) {
             displayWidth = info.imageWidth;
@@ -818,14 +1307,44 @@ PYBIND11_MODULE(imviz, m) {
             displayHeight = info.imageHeight;
         }
 
-        ImGui::Image(
-                (void*)(intptr_t)info.textureId,
-                ImVec2(displayWidth, displayHeight));
+        ImVec4 bc = interpretColor(borderCol);
+        ImVec4 tn = interpretColor(tint);
+        if (tn.w < 0) {
+            tn = ImVec4(1, 1, 1, 1);
+        }
+
+        // calculate expected bounding box beforehand
+        ImVec2 size(displayWidth, displayHeight);
+
+        // essentially copied from ImGui::Image function
+        ImGuiWindow* w = ImGui::GetCurrentWindow();
+        ImRect bb(w->DC.CursorPos, w->DC.CursorPos + size);
+        if (bc.w > 0.0f)
+            bb.Max += ImVec2(2, 2);
+
+        // upload to gpu
+
+        GLuint textureId = 0;
+
+        if (ImGui::IsRectVisible(bb.Min, bb.Max)) {
+            // only upload the image to gpu, if it's actually visible
+            // this improves performance for e.g. large lists of images
+            textureId = uploadImage(id, info, image);
+        }
+
+        ImGui::Image((void*)(intptr_t)textureId,
+                     size,
+                     ImVec2(0, 0),
+                     ImVec2(1, 1),
+                     tn,
+                     bc);
     },
     py::arg("id"),
     py::arg("image"),
     py::arg("width") = -1,
-    py::arg("height") = -1);
+    py::arg("height") = -1,
+    py::arg("tint") = py::array(),
+    py::arg("border_col") = py::array());
 
     m.def("plot_image", [&](
                 std::string id,
@@ -835,10 +1354,8 @@ PYBIND11_MODULE(imviz, m) {
                 double displayWidth,
                 double displayHeight) {
 
-        ImageInfo info = processImage(id, image);
-
-        // display texture
-
+        ImageInfo info = interpretImage(image);
+        
         if (displayWidth < 0) {
             displayWidth = info.imageWidth;
         }
@@ -846,12 +1363,14 @@ PYBIND11_MODULE(imviz, m) {
             displayHeight = info.imageHeight;
         }
 
+        GLuint textureId = uploadImage(id, info, image);
+
         ImPlotPoint boundsMin(x, y);
         ImPlotPoint boundsMax(x + displayWidth, y + displayHeight);
 
         ImPlot::PlotImage(
                 id.c_str(),
-                (void*)(intptr_t)info.textureId,
+                (void*)(intptr_t)textureId,
                 boundsMin,
                 boundsMax);
     },
@@ -975,6 +1494,8 @@ PYBIND11_MODULE(imviz, m) {
                       float markerSize,
                       float markerWeight) {
 
+        // interpret marker format
+
         std::smatch match;
         std::regex_search(fmt, match, viz.re);
 
@@ -987,10 +1508,6 @@ PYBIND11_MODULE(imviz, m) {
             }
         }
 
-        ImPlot::PushStyleVar(ImPlotStyleVar_LineWeight, lineWeight);
-        ImPlot::PushStyleVar(ImPlotStyleVar_MarkerSize, markerSize);
-        ImPlot::PushStyleVar(ImPlotStyleVar_MarkerWeight, markerWeight);
-
         if (groups[2] == "o") {
             ImPlot::PushStyleVar(ImPlotStyleVar_Marker, ImPlotMarker_Circle);
         } else if (groups[2] == "s") {
@@ -1005,46 +1522,27 @@ PYBIND11_MODULE(imviz, m) {
             ImPlot::PushStyleVar(ImPlotStyleVar_Marker, ImPlotMarker_None);
         }
 
-        size_t yCount = y.shape()[0];
+        // set style vars
 
-        std::vector<double> indices;
-        const double* xDataPtr = nullptr;
-        const double* yDataPtr = nullptr;
-        size_t count = 0;
+        ImPlot::PushStyleVar(ImPlotStyleVar_LineWeight, lineWeight);
+        ImPlot::PushStyleVar(ImPlotStyleVar_MarkerSize, markerSize);
+        ImPlot::PushStyleVar(ImPlotStyleVar_MarkerWeight, markerWeight);
 
-        if (1 == x.ndim() && 0 == yCount) {
-            count = x.shape()[0];
-            indices.resize(count);
-            for (size_t i = 0; i < count; ++i) {
-                indices[i] = i;
-            }
-            xDataPtr = indices.data();
-            yDataPtr = x.data();
-        } else if (2 == x.ndim() && 0 == yCount) {
-            size_t len0 = x.shape()[0];
-            size_t len1 = x.shape()[1];
-            if (len0 == 2) {
-                xDataPtr = x.data();
-                yDataPtr = x.data() + len1;
-                count = len1;
-            }
-        } else if (1 == x.ndim() && 1 == y.ndim()) {
-            count = std::min(x.shape()[0], y.shape()[0]);
-            xDataPtr = x.data();
-            yDataPtr = y.data();
-        } 
+        // interpret data
 
-        if (count == 0) {
-            throw std::runtime_error("Data format could not be interpreted");
-        }
+        PlotArrayInfo pai = interpretPlotArrays(x, y);
+
+        // plot lines and markers
 
         if (groups[1] == "-") {
-            ImPlot::PlotLine(label.c_str(), xDataPtr, yDataPtr, count);
+            ImPlot::PlotLine(label.c_str(), pai.xDataPtr, pai.yDataPtr, pai.count);
         } else {
-            ImPlot::PlotScatter(label.c_str(), xDataPtr, yDataPtr, count);
+            ImPlot::PlotScatter(label.c_str(), pai.xDataPtr, pai.yDataPtr, pai.count);
         }
 
-        size_t shadeCount = std::min(count, (size_t)shade.shape()[0]);
+        // plot shade if needed
+
+        size_t shadeCount = std::min(pai.count, (size_t)shade.shape()[0]);
 
         if (shadeCount != 0) {
             if (1 == shade.ndim()) {
@@ -1052,7 +1550,11 @@ PYBIND11_MODULE(imviz, m) {
                 auto mean = py::cast<py::array_t<double>>(y[py::slice(0, shadeCount, 1)]);
                 py::array_t<double> upper = mean + shade;
                 py::array_t<double> lower = mean - shade;
-                ImPlot::PlotShaded(label.c_str(), xDataPtr, lower.data(), upper.data(), shadeCount);
+                ImPlot::PlotShaded(label.c_str(),
+                                   pai.xDataPtr,
+                                   lower.data(),
+                                   upper.data(),
+                                   shadeCount);
                 ImPlot::PopStyleVar();
             }
         }
@@ -1069,6 +1571,40 @@ PYBIND11_MODULE(imviz, m) {
     py::arg("marker_size") = 4.0f, 
     py::arg("marker_weight") = 1.0f);
 
+    m.def("plot_bars", [&](array_like<double> x,
+                           array_like<double> y,
+                           std::string label,
+                           double width,
+                           double shift,
+                           bool horizontal) {
+
+        PlotArrayInfo pai = interpretPlotArrays(x, y);
+
+        if (horizontal) {
+            ImPlot::PlotBarsH(
+                    label.c_str(),
+                    pai.xDataPtr,
+                    pai.yDataPtr,
+                    pai.count,
+                    width,
+                    shift);
+        } else {
+            ImPlot::PlotBars(
+                    label.c_str(),
+                    pai.xDataPtr,
+                    pai.yDataPtr,
+                    pai.count,
+                    width,
+                    shift);
+        }
+    },
+    py::arg("x"),
+    py::arg("y") = py::array(),
+    py::arg("label") = "",
+    py::arg("width") = 0.5,
+    py::arg("shift") = 0.0,
+    py::arg("horizontal") = false);
+
     m.def("drag_point", [&](std::string label,
                             array_like<double> point,
                             bool showLabel,
@@ -1091,6 +1627,42 @@ PYBIND11_MODULE(imviz, m) {
     py::arg("color") = py::array_t<double>(),
     py::arg("radius") = 4.0);
 
+    m.def("plot_vlines", [&](std::string label,
+                            array_like<double> xs,
+                            array_like<double> color,
+                            float width) {
+
+        assert_shape(xs, {{-1}});
+
+        ImVec4 c = interpretColor(color);
+
+        ImPlot::SetNextLineStyle(c, width);
+
+        ImPlot::PlotVLines(label.c_str(), xs.data(), xs.shape(0));
+    },
+    py::arg("label"),
+    py::arg("xs"),
+    py::arg("color") = py::array_t<double>(),
+    py::arg("width") = 1.0);
+
+    m.def("plot_hlines", [&](std::string label,
+                            array_like<double> ys,
+                            array_like<double> color,
+                            float width) {
+
+        assert_shape(ys, {{-1}});
+
+        ImVec4 c = interpretColor(color);
+
+        ImPlot::SetNextLineStyle(c, width);
+
+        ImPlot::PlotHLines(label.c_str(), ys.data(), ys.shape(0));
+    },
+    py::arg("label"),
+    py::arg("ys"),
+    py::arg("color") = py::array_t<double>(),
+    py::arg("width") = 1.0);
+
     m.def("drag_vline", [&](std::string label,
                             double x,
                             bool showLabel,
@@ -1129,6 +1701,29 @@ PYBIND11_MODULE(imviz, m) {
     py::arg("color") = py::array_t<double>(),
     py::arg("width") = 1.0);
 
+    m.def("get_plot_pos", ImPlot::GetPlotPos);
+    m.def("get_plot_size", ImPlot::GetPlotSize);
+
+    m.def("get_plot_query", [&]() {
+        return ImPlot::GetPlotQuery();
+    });
+
+    m.def("set_plot_query", [&](ImPlotLimits query) {
+        ImPlot::SetPlotQuery(query);
+    });
+
+    m.def("get_plot_limits", [&]() {
+        return ImPlot::GetPlotLimits();
+    });
+
+    m.def("get_plot_mouse_pos", [&]() {
+        return ImPlot::GetPlotMousePos();
+    });
+
+    m.def("plot_contains", [&](ImPlotPoint point) {
+        return ImPlot::GetPlotLimits().Contains(point.x, point.y);
+    });
+
     m.def("annotate", [&](
                 double x,
                 double y,
@@ -1163,6 +1758,80 @@ PYBIND11_MODULE(imviz, m) {
     py::arg("offset") = py::array(),
     py::arg("clamp") = false);
 
+    m.def("begin_popup_context_item", [&](std::string label) {
+
+        return ImGui::BeginPopupContextItem(
+                label.empty() ? 0 : label.c_str());
+    },
+    py::arg("label") = "");
+
+    m.def("begin_popup", [&](std::string label) {
+
+        return ImGui::BeginPopup(label.c_str());
+    });
+
+    m.def("file_dialog_popup", [&](std::string label, std::string path, std::string confirmText) {
+
+        bool mod = ImGui::FileDialogPopup(label.c_str(), confirmText.c_str(), path);
+        viz.setMod(mod);
+
+        return path;
+    },
+    py::arg("label"),
+    py::arg("path"),
+    py::arg("confirmText") = "Ok");
+
+    m.def("begin_popup_modal", [&](std::string label) {
+
+        return ImGui::BeginPopupModal(
+                label.c_str(),
+                NULL,
+                ImGuiWindowFlags_AlwaysAutoResize);
+    });
+
+    m.def("open_popup", [&](std::string label) {
+
+        return ImGui::OpenPopup(label.c_str());
+    });
+
+    m.def("close_current_popup", ImGui::CloseCurrentPopup);
+    m.def("end_popup", ImGui::EndPopup);
+
+    m.def("get_id", [&](std::string id) {
+        return ImGui::GetID(id.c_str());
+    },
+    py::arg("id"));
+
+    m.def("push_id", [&](std::string id) {
+        ImGui::PushID(id.c_str());
+    },
+    py::arg("id"));
+
+    m.def("pop_id", ImGui::PopID);
+
+    m.def("set_item_default_focus", ImGui::SetItemDefaultFocus);
+
+    m.def("separator", ImGui::Separator);
+
+    m.def("selectable", [&](std::string label, bool selected, ImVec2 size) { 
+
+        bool s = ImGui::Selectable(label.c_str(), selected, 0, size);
+
+        if (selected != s) {
+            viz.setMod(true);
+        }
+
+        return s;
+    },
+    py::arg("label"),
+    py::arg("selected"),
+    py::arg("size") = ImVec2(0, 0));
+
+    m.def("get_content_region_avail", ImGui::GetContentRegionAvail);
+
+    m.def("begin_tooltip", ImGui::BeginTooltip);
+    m.def("end_tooltip", ImGui::EndTooltip);
+
     m.def("activate_svg", [&]() {
 
         //ImDrawList::svg = new std::stringstream();
